"""Redis RepositoryType"""

import datetime as dt
import functools
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Self

import ormsgpack
import redis

from gbp_ps.exceptions import RecordAlreadyExists, RecordNotFoundError
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

ENCODING = "ascii"

dumps: Callable[[Any], bytes] = functools.partial(
    ormsgpack.packb, option=ormsgpack.OPT_NAIVE_UTC
)
loads: Callable[[bytes], Any] = ormsgpack.unpackb  # pylint: disable=no-member


@dataclass(kw_only=True, frozen=True, slots=True)
class Key:
    """Redis key bytes parsed"""

    redis_key: str
    machine: str
    build_id: str
    package: str

    def __bytes__(self) -> bytes:
        return (
            f"{self.redis_key}:{self.machine}:{self.package}:{self.build_id}"
        ).encode(ENCODING)

    @classmethod
    def from_bytes(cls, b: bytes) -> Self:
        """Return the redis key from bytes"""
        string = b.decode(ENCODING)
        redis_key, machine, package, build_id = string.split(":")

        return cls(
            redis_key=redis_key, machine=machine, build_id=build_id, package=package
        )

    @classmethod
    def from_process(cls, process: BuildProcess, redis_key: str) -> Self:
        """Return Key given BuildProcess and redis_key"""
        return cls(
            redis_key=redis_key,
            machine=process.machine,
            build_id=process.build_id,
            package=process.package,
        )


class RedisRepository:
    """Redis backend for the process table"""

    def __init__(self, settings: Settings) -> None:
        self._redis = redis.Redis.from_url(settings.REDIS_URL)
        self._key = settings.REDIS_KEY
        self.time = settings.REDIS_KEY_EXPIRATION

    def key(self, process: BuildProcess) -> bytes:
        """Return the redis key for the given BuildProcess"""
        return bytes(Key.from_process(process, self._key))

    def value(self, process: BuildProcess) -> bytes:
        """Return the redis value for the given BuildProcess"""
        return dumps((process.build_host, process.phase, process.start_time))

    def process_to_redis(self, process: BuildProcess) -> tuple[bytes, bytes]:
        """Return the redis key and value for the given BuildProcess"""
        return self.key(process), self.value(process)

    def redis_to_process(self, key_bytes: bytes, value: bytes) -> BuildProcess:
        """Convert the given key and value to a BuildProcess"""
        key = Key.from_bytes(key_bytes)
        data = loads(value)

        return BuildProcess(
            build_host=data[0],
            build_id=key.build_id,
            machine=key.machine,
            package=key.package,
            phase=data[1],
            start_time=dt.datetime.fromisoformat(data[2]),
        )

    def add_process(self, process: BuildProcess) -> None:
        """Add the given BuildProcess to the repository

        If the process already exists in the repo, RecordAlreadyExists is raised
        """
        # If this package exists in another build, remove it. This (usually) means the
        # other build failed
        self.delete_existing_processes(process)
        key, value = self.process_to_redis(process)
        previous = self._redis.get(key)

        if previous and self.redis_to_process(key, previous).is_same_as(process):
            raise RecordAlreadyExists(process)

        self._redis.setex(key, self.time, value)

    def delete_existing_processes(self, process: BuildProcess) -> int:
        """Delete existing processes like process

        By "existing" we mean processes in Redis that have the same machine and package
        but different build_id.

        Return the number of processes deleted.
        """
        build_id = process.build_id
        deleted_count = 0
        pattern = f"{self._key}:{process.machine}:{process.package}:*".encode(ENCODING)
        for key_bytes in self._redis.keys(pattern):
            key = Key.from_bytes(key_bytes)
            if key.build_id != build_id:
                value = self._redis.get(key_bytes)
                assert value
                existing_process = self.redis_to_process(key_bytes, value)
                if existing_process.phase in BuildProcess.build_phases:
                    self._redis.delete(key_bytes)
                    deleted_count += 1
        return deleted_count

    def update_process(self, process: BuildProcess) -> None:
        """Update the given build process

        Only updates the phase field

        If the build process doesn't exist in the repo, RecordNotFoundError is raised.
        """
        key_bytes = self.key(process)
        previous_value = self._redis.get(key_bytes)

        if previous_value is None:
            raise RecordNotFoundError(process)

        self.redis_to_process(key_bytes, previous_value).ensure_updateable(process)
        new_value = (process.build_host, process.phase, loads(previous_value)[2])
        self._redis.setex(key_bytes, self.time, dumps(new_value))

    def get_processes(
        self, include_final: bool = False, machine: str | None = None
    ) -> Iterable[BuildProcess]:
        """Return the process records from the repository

        If include_final is True also include processes in their "final" phase. The
        default value is False.
        """
        processes = []

        for key_bytes in self._redis.keys(f"{self._key}:*".encode(ENCODING)):
            if machine and Key.from_bytes(key_bytes).machine != machine:
                continue
            if value := self._redis.get(key_bytes):
                process = self.redis_to_process(key_bytes, value)

                if include_final or not process.is_finished():
                    processes.append(process)

        processes.sort(key=lambda process: process.start_time)
        return processes



# pylint: disable=missing-docstring,too-few-public-methods,redefined-outer-name
import datetime as dt
from typing import Any
from unittest import mock

import factory
import gbp_testkit.fixtures as testkit
from django.test import TestCase as DjangoTestCase
from unittest_fixtures import FixtureContext, Fixtures, fixture

from gbp_ps.repository import Repo, RepositoryType, add_or_update_process, sqlite
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

LOCAL_TIMEZONE = dt.timezone(dt.timedelta(days=-1, seconds=61200), "PDT")
PACKAGES = (
    "media-libs/tiff-4.7.0",
    "app-misc/pax-utils-1.3.8",
    "media-libs/x265-3.6-r1",
    "sys-fs/cryptsetup-2.7.5-r1",
    "sys-devel/gcc-14.2.1_p20240921",
    "sys-fs/cryptsetup-2.7.5",
)


class TestCase(DjangoTestCase):
    """Custom TestCase for gbp-ps tests"""


class BuildProcessFactory(factory.Factory):
    class Meta:
        model = BuildProcess

    machine = "babette"
    build_id = factory.Sequence(str)
    build_host = "builder"
    package = factory.Iterator(PACKAGES)
    phase = factory.Iterator(BuildProcess.build_phases)
    start_time = factory.LazyFunction(
        lambda: dt.datetime.now(tz=dt.UTC).replace(microsecond=0)
    )


def make_build_process(**kwargs: Any) -> BuildProcess:
    """Create (and save) a BuildProcess"""
    settings = Settings.from_environ()
    add_to_repo = kwargs.pop("add_to_repo", True)
    update_repo = kwargs.pop("update_repo", False)
    attrs: dict[str, Any] = {
        "build_host": "jenkins",
        "build_id": "1031",
        "machine": "babette",
        "package": "sys-apps/systemd-254.5-r1",
        "phase": "compile",
        "start_time": dt.datetime(2023, 11, 11, 12, 20, 52, tzinfo=dt.timezone.utc),
    }
    attrs.update(**kwargs)
    build_process = BuildProcess(**attrs)

    if add_to_repo:
        repo = Repo(settings)
        if update_repo:
            add_or_update_process(repo, build_process)
        else:
            repo.add_process(build_process)

    return build_process


@fixture()
def build_process(_fixtures: Fixtures, **options: Any) -> BuildProcess:
    return BuildProcessFactory(**options)


@fixture(testkit.tmpdir)
def tempdb(fixtures: Fixtures) -> str:
    return f"{fixtures.tmpdir}/processes.db"


@fixture(testkit.tmpdir)
def environ(
    fixtures: Fixtures,
    environ: dict[str, str] | None = None,  # pylint: disable=redefined-outer-name
) -> FixtureContext[dict[str, str]]:
    new_environ: dict[str, str] = next(testkit.environ(fixtures), {}).copy()
    new_environ["GBP_PS_SQLITE_DATABASE"] = f"{fixtures.tmpdir}/db.sqlite"
    new_environ.update(environ or {})
    with mock.patch.dict("os.environ", new_environ):
        yield new_environ


@fixture(environ)
def settings(_fixtures: Fixtures) -> Settings:
    return Settings.from_environ()


@fixture(settings)
def repo(fixtures: Fixtures) -> RepositoryType:
    return Repo(fixtures.settings)


@fixture(tempdb)
def repo_fixture(fixtures: Fixtures) -> sqlite.SqliteRepository:
    return sqlite.SqliteRepository(Settings(SQLITE_DATABASE=fixtures.tempdb))


@fixture()
def local_timezone(
    _: Fixtures, local_timezone: dt.timezone = LOCAL_TIMEZONE
) -> FixtureContext[dt.timezone]:
    with mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=local_timezone):
        yield local_timezone


@fixture()
def get_today(
    _: Fixtures, get_today: dt.date = dt.date(2023, 11, 11)
) -> FixtureContext[dt.date]:
    with mock.patch("gbp_ps.cli.ps.utils.get_today", new=lambda: get_today):
        yield get_today


@fixture()
def now(
    _: Fixtures,
    now: dt.datetime = dt.datetime(2023, 11, 11, 16, 30, tzinfo=LOCAL_TIMEZONE),
) -> FixtureContext[dt.datetime]:
    with mock.patch("gbp_ps.utils.now") as mock_now:
        mock_now.return_value = now
        yield now



"""CLI unit tests for gbp-ps add-process subcommand"""

# pylint: disable=missing-docstring,unused-argument
import datetime as dt
import platform
from argparse import ArgumentParser
from unittest import mock

import gbp_testkit.fixtures as testkit
from gbp_testkit.helpers import parse_args
from unittest_fixtures import Fixtures, given

from gbp_ps.cli import add_process

from . import lib


@given(lib.repo, testkit.gbp, testkit.console)
class AddProcessTests(lib.TestCase):
    """Tests for gbp add-process"""

    maxDiff = None

    @mock.patch("gbp_ps.cli.add_process.now")
    def test(self, mock_now: mock.Mock, fixtures: Fixtures) -> None:
        now = mock_now.return_value = dt.datetime(2023, 11, 20, 17, 57, tzinfo=dt.UTC)
        proc = lib.make_build_process(
            add_to_repo=False, build_host=platform.node(), start_time=now
        )
        console = fixtures.console
        cmdline = f"gbp add-process {proc.machine} {proc.build_id} {proc.package} {proc.phase}"
        args = parse_args(cmdline)
        exit_status = add_process.handler(args, fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        self.assertEqual([*fixtures.repo.get_processes()], [proc])

    def test_parse_args(self, fixtures: Fixtures) -> None:
        # Just ensure that parse_args is there and works
        parser = ArgumentParser()
        add_process.parse_args(parser)


@given(lib.tempdb, lib.repo_fixture, process=lib.build_process)
class AddProcessAddLocalProcessesTests(lib.TestCase):
    def test(self, fixtures: Fixtures) -> None:
        process = fixtures.process

        add_process.add_local_process(fixtures.tempdb)(process)

        result = fixtures.repo.get_processes()

        self.assertEqual(list(result), [process])


@given(lib.build_process)
@mock.patch("gbp_ps.cli.add_process.now")
@mock.patch("gbp_ps.cli.add_process.platform.node")
class BuildProcessFromArgsTests(lib.TestCase):
    def test(self, node: mock.Mock, now: mock.Mock, fixtures: Fixtures) -> None:
        expected = fixtures.build_process
        node.return_value = expected.build_host
        now.return_value = expected.start_time
        cmdline = (
            f"gbp add-process {expected.machine} {expected.build_id} {expected.package}"
            f" {expected.phase}"
        )
        args = parse_args(cmdline)

        process = add_process.build_process_from_args(args)

        self.assertEqual(process, expected)



"""Mutation GraphQL type for gbp-ps"""

from typing import Any, TypeAlias

from ariadne import ObjectType, convert_kwargs_to_snake_case
from graphql import GraphQLResolveInfo

from gbp_ps.repository import Repo, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

Info: TypeAlias = GraphQLResolveInfo
Mutation = ObjectType("Mutation")

ADD_BUILD_FIELDS = {"machine", "id", "package", "phase"}


@Mutation.field("addBuildProcess")
@convert_kwargs_to_snake_case
def _(_obj: Any, _info: Info, process: dict[str, Any]) -> None:
    """Add the given process to the process table

    If the process already exists in the table, it is updated with the new value
    """
    # Don't bother when required fields are empty.
    if not all(process[field] for field in ADD_BUILD_FIELDS):
        return

    process["build_id"] = process.pop("id")
    add_or_update_process(Repo(Settings.from_environ()), BuildProcess(**process))



"""Tests for gbp_ps.utils"""

# pylint: disable=missing-docstring,unused-argument
import datetime as dt
from unittest import TestCase, mock

from unittest_fixtures import Fixtures, given

from gbp_ps import utils

from . import lib


@given(lib.local_timezone)
class GetTodayTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        now = dt.datetime(2024, 2, 7, 20, 10, 57, 312885)
        with mock.patch.object(utils, "now", return_value=now):
            expected = dt.date(2024, 2, 7)
            self.assertEqual(expected, utils.get_today())


class FormatTimestampTests(TestCase):

    def test_when_today(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10)
        today = timestamp.date()

        with mock.patch("gbp_ps.utils.get_today", return_value=today):
            date_str = utils.format_timestamp(timestamp)

        self.assertEqual(date_str, "[timestamp]20:10:00[/timestamp]")

    def test_when_not_today(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10)
        today = (timestamp + dt.timedelta(hours=24)).date()

        with mock.patch("gbp_ps.utils.get_today", return_value=today):
            date_str = utils.format_timestamp(timestamp)

        self.assertEqual(date_str, "[timestamp]Feb07[/timestamp]")


class FormatElapsedTests(TestCase):
    def test(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10, 37)
        since = dt.datetime(2024, 2, 7, 20, 14, 51)

        date_str = utils.format_elapsed(timestamp, since)

        self.assertEqual("[timestamp]0:04:14[/timestamp]", date_str)

    def test_with_default_since(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10, 37)
        since = dt.datetime(2024, 2, 7, 20, 14, 51)

        with mock.patch("gbp_ps.utils.now", return_value=since):
            date_str = utils.format_elapsed(timestamp)

        self.assertEqual("[timestamp]0:04:14[/timestamp]", date_str)


class FindTests(TestCase):
    def test_item_in_sequence(self) -> None:
        self.assertEqual(1, utils.find("test", ("this", "test", "is", "a")))

    def test_item_not_insequence(self) -> None:
        self.assertEqual(-1, utils.find("test", ("a", "is", "this")))

    def test_empty_sequence_self(self) -> None:
        self.assertEqual(-1, utils.find("test", ()))

    def test_item_exists_twice_in_sequence(self) -> None:
        self.assertEqual(1, utils.find("test", ("is", "test", "a", "this", "test")))



"""BuildProcess GraphQL resolver for gbp-ps"""

from typing import TypeAlias

from ariadne import ObjectType
from graphql import GraphQLResolveInfo

from gbp_ps import types

BuildProcess = ObjectType("BuildProcess")
Info: TypeAlias = GraphQLResolveInfo

# pylint: disable=missing-docstring


@BuildProcess.field("id")
def _(process: types.BuildProcess, _info: Info) -> str:
    return process.build_id



"""GraphQL interface for gbp-ps"""

from importlib import resources

from ariadne import gql

from .build_process import BuildProcess
from .mutations import Mutation
from .queries import Query

type_defs = gql(resources.read_text("gbp_ps.graphql", "schema.graphql"))
resolvers = [BuildProcess, Mutation, Query]



"""Query GraphQL resolver for gbp-ps"""

from typing import Any, Iterable, TypeAlias

from ariadne import ObjectType, convert_kwargs_to_snake_case
from graphql import GraphQLResolveInfo

from gbp_ps.repository import Repo
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

Info: TypeAlias = GraphQLResolveInfo
Query = ObjectType("Query")


get_processes = Repo(Settings.from_environ()).get_processes


@Query.field("buildProcesses")
@convert_kwargs_to_snake_case
def _(
    _obj: Any, _info: Info, *, include_final: bool = False, machine: str
) -> Iterable[BuildProcess]:
    """Return the list of BuildProcesses

    If include_final is True also include processes in their "final" phase. The default
    value is False.
    """
    return get_processes(include_final=include_final, machine=machine)



"""Tests for gbp-ps repositories"""

# pylint: disable=missing-docstring, duplicate-code
from dataclasses import replace
from typing import Any, Callable
from unittest import mock

import fakeredis
from unittest_fixtures import Fixtures, given, parametrized, where

from gbp_ps.exceptions import (
    RecordAlreadyExists,
    RecordNotFoundError,
    UpdateNotAllowedError,
)
from gbp_ps.repository import Repo, RepositoryType, add_or_update_process, sqlite
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

from . import lib

HOST = 0
REDIS_FROM_URL = "gbp_ps.repository.redis.redis.Redis.from_url"


def get_repo(backend: str, settings: Settings) -> RepositoryType:
    global HOST  # pylint: disable=global-statement
    settings = replace(settings, STORAGE_BACKEND=backend)
    if backend == "redis":
        fake_redis = fakeredis.FakeRedis(host=f"host{(HOST := HOST + 1)}")
        repo_patch = mock.patch(REDIS_FROM_URL, return_value=fake_redis)
    else:
        repo_patch = mock.MagicMock()

    with repo_patch:
        return Repo(settings)


def repos(*names: str) -> Callable[[Callable[[Any, str], None]], None]:
    return parametrized([[name] for name in names])


@given(lib.environ, lib.settings, lib.build_process)
@where(
    environ={
        "GBP_PS_KEY_EXPIRATION": "3600",
        "GBP_PS_REDIS_KEY": "gbp-ps-test",
        "GBP_PS_STORAGE_BACKEND": "sqlite",
    },
    build_process__phase="compile",
)
class RepositoryTests(lib.TestCase):
    @repos("django", "redis", "sqlite")
    def test_add_process(self, backend: str, fixtures: Fixtures) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = fixtures.build_process
        repo.add_process(build_process)
        self.assertEqual([*repo.get_processes()], [build_process])

    @repos("django", "redis", "sqlite")
    def test_add_process_when_already_exists(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = fixtures.build_process
        repo.add_process(build_process)

        with self.assertRaises(RecordAlreadyExists):
            repo.add_process(build_process)

    @repos("django", "redis", "sqlite")
    def test_add_process_same_package_in_different_builds_exist_only_once(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        dead_process: BuildProcess = fixtures.build_process
        repo.add_process(dead_process)
        new_process = replace(
            dead_process, build_id=str(int(dead_process.build_id) + 1)
        )
        repo.add_process(new_process)

        self.assertEqual([*repo.get_processes()], [new_process])

    @repos("django", "redis", "sqlite")
    def test_update_process(self, backend: str, fixtures: Fixtures) -> None:
        repo = get_repo(backend, fixtures.settings)
        orig_process: BuildProcess = replace(fixtures.build_process, phase="compile")
        repo.add_process(orig_process)

        updated_process = replace(orig_process, phase="postinst")

        repo.update_process(updated_process)

        expected = replace(orig_process, phase="postinst")
        self.assertEqual([*repo.get_processes()], [expected])

    @repos("django", "redis", "sqlite")
    def test_update_process_finalize_when_not_owned(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        # This demonstrates the concept of build host "ownership". A a process can only
        # be updated with a "final" phase if the build host is the same. Otherwise it
        # should raise an exception
        repo = get_repo(backend, fixtures.settings)
        process1 = lib.make_build_process(add_to_repo=False)
        repo.add_process(process1)
        process2 = replace(process1, build_host="badhost", phase="clean")

        with self.assertRaises(UpdateNotAllowedError):
            repo.update_process(process2)

    @repos("django", "redis", "sqlite")
    def test_new_build_does_not_wipe_out_previous_if_no_packages(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        # We want to demonstrate the following scenario
        #
        # | t | machine | build | package  | phase |
        # |---|---------|-------|----------|-------|
        # | 1 | babette |     1 | pipeline | clean |
        # | 2 | babette |     2 | pipeline | world |
        # | 3 | babette |     1 | pipeline | index |
        #
        # In this scenario, the update in t3 should not wipe out t2
        repo = get_repo(backend, fixtures.settings)

        t1 = lib.make_build_process(
            machine="babette",
            build_id="1",
            package="pipeline",
            phase="clean",
            add_to_repo=False,
        )
        t2 = lib.make_build_process(
            machine="babette",
            build_id="2",
            package="pipeline",
            phase="world",
            add_to_repo=False,
        )
        t3 = lib.make_build_process(
            machine="babette",
            build_id="1",
            package="pipeline",
            phase="index",
            add_to_repo=False,
        )
        add_or_update_process(repo, t1)
        add_or_update_process(repo, t2)
        add_or_update_process(repo, t3)

        processes = set(repo.get_processes())

        self.assertEqual({t2, t3}, processes)

    @repos("django", "redis")
    def test_add_or_update_process_can_handle_buildhost_changes(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        orig_process: BuildProcess = replace(fixtures.build_process, phase="clean")
        repo.add_process(orig_process)

        updated_process = replace(orig_process, build_host="gbp", phase="pull")

        add_or_update_process(repo, updated_process)

        expected = replace(orig_process, build_host="gbp", phase="pull")
        self.assertEqual([*repo.get_processes()], [expected])

    @repos("django", "redis", "sqlite")
    def test_add_or_update_ignores_notallowederror(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        process1 = lib.make_build_process(add_to_repo=False)
        repo.add_process(process1)
        process2 = replace(process1, build_host="badhost", phase="clean")

        add_or_update_process(repo, process2)

        self.assertEqual([*repo.get_processes()], [process1])

    @repos("django", "redis", "sqlite")
    def test_update_process_when_process_not_in_db(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = fixtures.build_process

        with self.assertRaises(RecordNotFoundError):
            repo.update_process(build_process)

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_empty_list(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        self.assertEqual([*repo.get_processes()], [])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_process(self, backend: str, fixtures: Fixtures) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = fixtures.build_process
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes()], [build_process])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_final_process(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = replace(fixtures.build_process, phase="postrm")
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes()], [])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_include_final_process(
        self, backend: str, fixtures: Fixtures
    ) -> None:
        repo = get_repo(backend, fixtures.settings)
        build_process: BuildProcess = replace(fixtures.build_process, phase="postrm")
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes(include_final=True)], [build_process])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_machine(self, backend: str, fixtures: Fixtures) -> None:
        repo = get_repo(backend, fixtures.settings)
        repo.add_process(fixtures.build_process)
        build_process = replace(fixtures.build_process, machine="laika")
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes(machine="laika")], [build_process])

    def test_repo_factory_success(self, fixtures: Fixtures) -> None:
        settings = replace(fixtures.settings, STORAGE_BACKEND="sqlite")
        repo = Repo(settings)

        self.assertTrue(isinstance(repo, sqlite.SqliteRepository))

    def test_repo_factory_failure(self, fixtures: Fixtures) -> None:
        settings = replace(fixtures.settings, STORAGE_BACKEND="bogus")

        with self.assertRaises(ValueError):
            Repo(settings)



#!/usr/bin/env python
"""Run tests for Gentoo Build Publisher"""
import argparse
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


def main() -> None:
    """Program entry point"""
    args = parse_args()
    os.environ["DJANGO_SETTINGS_MODULE"] = args.settings

    # These values are required in order to import the publisher module
    os.environ.setdefault("BUILD_PUBLISHER_JENKINS_BASE_URL", "http://jenkins.invalid/")
    os.environ.setdefault("BUILD_PUBLISHER_STORAGE_PATH", "__testing__")

    django.setup()

    TestRunner = get_runner(settings)  # pylint: disable=invalid-name
    verbosity = 2 if args.verbose else 1
    test_runner = TestRunner(failfast=args.failfast, verbosity=verbosity)
    failures = test_runner.run_tests(args.tests)

    sys.exit(bool(failures))


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments"""
    default_settings = os.environ.get("DJANGO_SETTINGS_MODULE", "gbp_testkit.settings")
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--failfast", action="store_true", default=False)
    parser.add_argument("--settings", default=default_settings)
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument("tests", nargs="*", default=["tests"])

    return parser.parse_args()


if __name__ == "__main__":
    main()



"""gbp-ps"""

import importlib.metadata

__version__ = importlib.metadata.version("gbp-ps")

# Plugin definition
plugin = {
    "name": "gbp-ps",
    "version": __version__,
    "description": "A plugin to display your Gentoo Build Publisher processes",
    "app": "gbp_ps.django.gbp_ps",
    "graphql": "gbp_ps.graphql",
}
